"""
watcher.py - Defines watcher classes.

Every watcher class has a function 'watch()' that monitors some resource and
returns when it detects the state is changed. The target resources monitored
and their reaction vary by watcher classes, and users may choose one watcher
class they want to use through the command line argument of the coy program.

[For developer] All watcher classes must inherit the base class 'Watcher' and
override 'watch()' method with their own functionalities implemented. The
method must implement the same interface, meaning that it takes the same
arguments and returns an integer. The method should alleviate bursty changes
and small oscillation near a threshold in the monitored resource using, for
example, multi-stage thresholds, counters, and some other statistical methods.
If the method takes a long time to process, it should utilize 'proc.poll()' to
check to see if the target process dies during the processing, in which case
the result of the method is no longer needed and so the method may exit.
     If a class requires additional modules, the modules must be imported
inside the class declaration so that users can make use of the coy tool with
the minimum set of the prerequisite libraries.

Copyright 2018 coronocoya.net <nshou@coronocoya.net>
This code can be distributed under the terms of MIT License. Please see the
file COPYING for the license description.

"""

# Register the names of class here to activate.
ALL = ['DebuggingWatcher', 'CPUUsageWatcher']


class Watcher:
    """
    A base class of all watcher classes.

    It does nothing and must not be instantiated.

    Args & attributes:
      proc (subprocess.Popen): Forked process.

    """
    def __init__(self, proc=None):
        self.proc = proc

    def watch(self, *args, **kwargs):
        """
        Does nothing and immediately returns 0.

        Returns:
          int: Always 0.

        """
        return 0

    def set_proc(self, proc):
        """
        Stores a Popen object.

        Args:
          proc (subprocess.Popen): Forked process.

        Returns:
          subprocess.Popen: A Popen object previously set.

        """
        prev = self.proc
        self.proc = proc
        return prev


class DebuggingWatcher(Watcher):
    """
    A watcher class only for test use.

    It reads from stdin and returns the input as an integer.

    """
    def watch(self, *args, **kwargs):
        """
        Waits for input from stdin and returns it as an integer.

        Returns:
          int: Value provided by stdin

        """
        try:
            i = int(input('[DebuggingWatcher] '))
        except ValueError:
            i = 0

        return i


class CPUUsageWatcher(Watcher):
    """
    Watches if the other process begins a high-load task on CPU.

    CPUUsageWatcher monitors the system-wide CPU utilization as a percentage
    regularly (every 1/3 seconds by default) and calculates its EWMA. The other
    processes' load (OPL) is defined by subtracting the utilization of the
    target process alone (self.proc) from the EWMA. The function 'watch()'
    returns when OPL crosses the thresholds.
         EWMA is calculated by 'EWMA_n = a * U + (1 - a) * EWMA_n-1' where 'a'
    is 0.35 by default, 'U' is the last sample value of CPU usage.
         The watcher has 4 thresholds, th_max, th_high, th_low, and th_min. The
    default values of th_max and th_min are 50 and 25 respectively. Th_high is
    defined by 0.9 * th_max and th_low is 1.2 * th_min. If the OPL exceeds
    th_max or is less than th_min, 'watch()' immediately returns indicating so.
    It also returns when the OPL is 'th_max > OPL > th_high' for several times
    (2 * 'reciprocal for the sampling frequency') or 'th_low > OPL > th_min'
    for a certain times (1 * 'reciprocal for the sampling frequency'.)

    """
    def __init__(self, proc=None):
        super(CPUUsageWatcher, self).__init__(proc)
        self.interval = 1 / 3
        self.alpha = 0.35
        self.th_max = 50
        self.th_min = 25
        self.th_high = 0.9 * self.th_max
        self.th_low = 1.2 * self.th_min
        self.cnt_high = int(2 / self.interval)
        self.cnt_low = int(1 / self.interval)
        self.lastreport = 0 # 0:idle, 1:busy
        # TODO: Config parsing
        # TODO: Handle errors
        #   psutil import error check, config parse error, exceptions raised by
        #   psutil.Process, errors on unacceptable value range, etc..
        # TODO: Documentation on exceptions

    def watch(self, *args, **kwargs):
        """
        Monitors the CPU utilization and detects the state change of whether
        the system is busy or idle.

        Returns:
          int: 1 if the system turns busy from being idle, or 0 on the other
               way around.

        """
        import psutil
        counter_high = 0
        counter_low = 0
        targetp = psutil.Process(self.proc.pid)
        ncores = psutil.cpu_count()
        targetp.cpu_percent() # Ignore the first call

        ewma = psutil.cpu_percent(interval=self.interval) # EWMA_0
        while self.proc.poll() is None:
            opl = ewma - targetp.cpu_percent() / ncores

            if opl >= self.th_max:
                counter_high += 1
                counter_low = 0
                if self.lastreport == 0:
                    self.lastreport = 1
                    return 1 # idle -> busy

            elif self.th_max > opl >= self.th_high:
                counter_high += 1
                counter_low = 0
                if self.lastreport == 0 and counter_high >= self.cnt_high:
                    self.lastreport = 1
                    return 1 # idle -> busy

            elif self.th_low >= opl > self.th_min:
                counter_high = 0
                counter_low += 1
                if self.lastreport == 1 and counter_low >= self.cnt_low:
                    self.lastreport = 0
                    return 0 # busy -> idle

            elif self.th_min >= opl:
                counter_high = 0
                counter_low += 1
                if self.lastreport == 1:
                    self.lastreport = 0
                    return 0 # busy -> idle

            ewma = self.alpha * psutil.cpu_percent(interval=self.interval) + (1 - self.alpha) * ewma

        return 0 # Target process terminated or some error occurred.
        # TODO: catch exceptions raised by psutil



"""
coy.py - A program to control the behavior of process dynamically.

Coy monitors various kinds of resources ("watch") and affects the behavior of
the specified process ("control"). A 'some_task' process is controlled by coy
in the following example:

    $ python coy.py ./some_task

[For developer] The diagram below illustrates the control flow of the coy
program. Watcher.py and controller.py provide several watcher classes and
controller classes respectively each of which has its own features of how to
achieve watching/controlling.

          coy
           |
           +--> (fork process) --+
           |                     |
           v                     |
    +--> watcher                 |
    |      |                     |
    |      v                     |
    |    (detect state change)   |
    |      |                     |
    |      v                     |
    |    controller -----------> @ (effect on process)
    |      |                     |
    +----- @ (loop while         |
           |  process is alive)  |
           |                     |
           v                     |
           +<--------------------+
           |
           v
          end

Copyright 2018 coronocoya.net <nshou@coronocoya.net>
This code can be distributed under the terms of MIT License. Please see the
file COPYING for the license description.

"""

__version__ = '0.0.1'


import argparse
import sys
import subprocess
from . import watcher
from . import controller


def _print_error(message):
    """
    Prints out message prepended with a tag to stderr.

    Args:
      message (str): A string to be printed.

    """
    print('[coy] %s' % message, file=sys.stderr)


def _parse_command():
    """
    Parses command line. This function may exit and never return.

    Returns:
      argparse.Namespace: Parsed results.

    """
    parser = argparse.ArgumentParser(description='A program to control the behavior of process dynamically.', add_help=False)
    parser.add_argument('-w', '--watcher', help='A watcher to be used. Default is CPUUsageWatcher.', default='CPUUsageWatcher')
    parser.add_argument('-o', '--watcher-option', help='Options for watcher.')
    parser.add_argument('-c', '--controller', help='A controller to be used. Default is StopContController.', default='StopContController')
    parser.add_argument('-p', '--controller-option', help='Options for controller.')
    parser.add_argument('-h', '--help', help='Print this message or help message for watcher/controller specified by NAME, then exit.', nargs='?', const='self', metavar='NAME')
    parser.add_argument('-l', '--list', help='Print available watchers/controllers and exit.', action='store_true')
    parser.add_argument('-v', '--version', action='version', version=__version__)
    parser.add_argument('command', nargs=argparse.REMAINDER, metavar='COMMAND [OPTIONS ...]')
    res = parser.parse_args()

    if res.help is not None:
        if res.help == 'self':
            parser.print_help()
            sys.exit(0)
        else:
            # TODO: print_help(res.help)
            sys.exit(0)

    if res.list:
        # TODO: print_list()
        sys.exit(0)

    if not res.command:
        parser.print_help()
        sys.exit(1)

    return res


def prepare_watcher(watcher_name):
    """
    Instantiates and returns watcher object specified by name.

    Args:
      watcher_name (str): Name of watcher class to be instantiated.

    Returns:
      watcher.Watcher: A watcher object if successful, None otherwise.

    """
    watcher_obj = None

    if watcher_name not in dir(watcher):
        _print_error('%s is not defined as a watcher.' % watcher_name)
    elif watcher_name not in watcher.ALL:
        _print_error('%s is not activated.' % watcher_name)
    else:
        try:
            watcher_obj = getattr(watcher, watcher_name)()
        except Exception as err:
            _print_error('Failed to instantiate %s: %s' % (watcher_name, err))

    return watcher_obj


def prepare_controller(controller_name):
    """
    Instantiates and returns controller object specified by name.

    Args:
      controller_name (str): Name of controller class to be instantiated.

    Returns:
      controller.Controller: A controller object if successful, None otherwise.

    """
    controller_obj = None

    if controller_name not in dir(controller):
        _print_error('%s is not defined as a controller.' % controller_name)
    elif controller_name not in controller.ALL:
        _print_error('%s is not activated.' % controller_name)
    else:
        try:
            controller_obj = getattr(controller, controller_name)()
        except Exception as err:
            _print_error('Failed to instantiate %s: %s' % (controller_name, err))

    return controller_obj


def start_process(args):
    """
    Forks and starts a process and returns it as subprocess.Popen object.

    Args:
      args (list): Argument list of the process (ex. ['ls', '-l'].)

    Returns:
      subprocess.Popen: Popen object if successful, None otherwise.

    """
    proc = None

    try:
        proc = subprocess.Popen(args)
    except (OSError, ValueError, subprocess.SubprocessError) as err:
        _print_error('Failed to start process: %s' % err)

    return proc


def coy_work(args, watcher_name, controller_name):
    """
    Starts the main routine of the coy program.

    It immediately returns when it fails to {fork process | make watcher object
    | make controller object}. It also returns when the forked process
    terminates.

    Args:
      args (list): Argument for 'start_process()'.
      watcher_name (str): Argument for 'prepare_watcher()'.
      controller_name (str): Argument for 'prepare_controller()'.

    """
    watcher_obj = prepare_watcher(watcher_name)
    if watcher_obj is None:
        return

    controller_obj = prepare_controller(controller_name)
    if controller_obj is None:
        return

    proc = start_process(args)
    if proc is None:
        return

    watcher_obj.set_proc(proc)
    controller_obj.set_proc(proc)

    while proc.poll() is None:
        state = watcher_obj.watch()
        controller_obj.control(state)


def main():
    """Called when coy is run as a command line script."""
    args = _parse_command()
    coy_work(args.command, args.watcher, args.controller)


if __name__ == '__main__':
    main()



"""
controller.py - Defines controller classes.

Every controller class has a function 'control()' that causes some influence
on the forked process. The effect the function creates vary by controller
classes, and so users may choose one controller class they want to use through
the command line argument of the coy program.

[For developer] All controller classes must inherit the base class 'Controller'
and override 'control()' method with their own functionalities implemented.
The method must implement the same interface, meaning that it takes the same
arguments and returns a boolean. The method is called by coy with an argument
of integer which is the return value of 'watch()' function. You may change the
behavior of the method based on the integer argument.
     If a class requires additional modules, the modules must be imported
inside the class declaration so that users can make use of the coy tool with
the minimum set of the prerequisite libraries.

Copyright 2018 coronocoya.net <nshou@coronocoya.net>
This code can be distributed under the terms of MIT License. Please see the
file COPYING for the license description.

"""

# Register the names of class here to activate.
ALL = ['DebuggingController', 'StopContController']


class Controller:
    """
    A base class of all controller classes.

    It does nothing and must not be instantiated.

    Args & attributes:
      proc (subprocess.Popen): Forked process.

    """
    def __init__(self, proc=None):
        self.proc = proc

    def control(self, state, *args, **kwargs):
        """
        Does nothing and immediately returns True.

        Args:
          state (int): State value returned by the last 'watch()'.

        Returns:
          bool: Always True.

        """
        return True

    def set_proc(self, proc):
        """
        Stores a Popen object.

        Args:
          proc (subprocess.Popen): Forked process.

        Returns:
          subprocess.Popen: A Popen object previously set.

        """
        prev = self.proc
        self.proc = proc
        return prev


class DebuggingController(Controller):
    """
    A controller class only for test use.

    It prints out the PID of the forked process and state, then returns.

    """
    def control(self, state, *args, **kwargs):
        """
        Just prints out the PID of the forked process and state.

        Args:
          state (int): State value returned by the last 'watch()'.

        Returns:
          bool: Always True.

        """
        print(self.proc.pid, state)
        return True


class StopContController(Controller):
    """
    Controls processes using SIGSTOP and SIGCONT.

    This controller enables the forked process (self.proc) to completely yield
    up its CPU usage quota to the other processes when triggered.

    """
    def control(self, state, *args, **kwargs):
        """
        Sends a SIGSTOP/SIGCONT signal based on the 'state' argument.

        The method is stateless, meaning that it does not store any variables
        with regard to the state of the forked process. It sends a signal based
        only on the 'state' no matter whether the process is already stopped or
        still running. This is reasonable because the behavior of processes
        against SIGSTOP and SIGCONT is idempotent in most systems.

        Args:
          state (int): 1 to SIGSTOP, 0 to SIGCONT.

        Returns:
          bool: True if successful, False otherwise.

        """
        import signal

        if state == 1:
            self.proc.send_signal(signal.SIGSTOP)
            return True
        if state == 0:
            self.proc.send_signal(signal.SIGCONT)
            return True

        return False
        # TODO: error handling, process existence check.



from setuptools import setup
import coy

with open('README.md') as f:
    readme = f.read()

setup(
    name='coy',
    version=coy.coy.__version__,
    packages=['coy'],
    description='A program to control the behavior of process dynamically.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/nshou/coy',
    author='nshou',
    author_email='nshou@coronocoya.net',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ],
    entry_points={
        'console_scripts': ['coy=coy:main'],
    },
)
